# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library to retrain image classifier models."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import logging
import os
import tempfile

import tensorflow.compat.v2 as tf
from tensorflow_examples.lite.model_maker.core.optimization import warmup
from tensorflow_examples.lite.model_maker.core.task import model_util
from tensorflow_hub.tools.make_image_classifier import make_image_classifier_lib as hub_lib

DEFAULT_DECAY_SAMPLES = 10000 * 256
DEFAULT_WARMUP_EPOCHS = 2


def add_params(hparams, **kwargs):
  param_dict = {k: v for k, v in kwargs.items() if v is not None}

  return hparams._replace(**param_dict)


class HParams(
    collections.namedtuple(
        "HParams", hub_lib.HParams._fields + ("warmup_steps", "model_dir"))):
  """The hyperparameters for make_image_classifier.

  train_epochs: Training will do this many iterations over the dataset.
  do_fine_tuning: If true, the Hub module is trained together with the
    classification layer on top.
  batch_size: Each training step samples a batch of this many images.
  learning_rate: Base learning rate when train batch size is 256. Linear to the
    batch size.
  dropout_rate: The fraction of the input units to drop, used in dropout layer.
  warmup_steps: Number of warmup steps for warmup schedule on learning rate.
  model_dir: The location of the model checkpoint files.
  """

  @classmethod
  def get_hparams(cls, **kwargs):
    """Gets the hyperparameters for `train_image_classifier_lib`."""
    hparams = get_default_hparams()
    return add_params(hparams, **kwargs)


def get_default_hparams():
  """Returns a fresh HParams object initialized to default values."""
  default_hub_hparams = hub_lib.get_default_hparams()
  as_dict = default_hub_hparams._asdict()
  as_dict.update(
      train_epochs=10,
      do_fine_tuning=False,
      batch_size=64,
      learning_rate=0.004,
      dropout_rate=0.2,
      warmup_steps=None,
      model_dir=tempfile.mkdtemp(),
  )
  default_hparams = HParams(**as_dict)
  return default_hparams


def create_optimizer(init_lr, num_decay_steps, num_warmup_steps):
  """Creates an optimizer with learning rate schedule."""
  # Leverages cosine decay of the learning rate.
  learning_rate_fn = tf.keras.experimental.CosineDecay(
      initial_learning_rate=init_lr, decay_steps=num_decay_steps, alpha=0.0)
  if num_warmup_steps:
    learning_rate_fn = warmup.WarmUp(
        initial_learning_rate=init_lr,
        decay_schedule_fn=learning_rate_fn,
        warmup_steps=num_warmup_steps)
  optimizer = tf.keras.optimizers.RMSprop(
      learning_rate=learning_rate_fn, rho=0.9, momentum=0.9, epsilon=0.001)

  return optimizer


def hub_train_model(model, hparams, train_ds, validation_ds, steps_per_epoch):
  """Trains model with the given data and hyperparameters.

  If using a DistributionStrategy, call this under its `.scope()`.
  Args:
    model: The tf.keras.Model from _build_model().
    hparams: A namedtuple of hyperparameters. This function expects
      .train_epochs: a Python integer with the number of passes over the
        training dataset;
      .learning_rate: a Python float forwarded to the optimizer;
      .momentum: a Python float forwarded to the optimizer;
      .batch_size: a Python integer, the number of examples returned by each
        call to the generators.
    train_ds: tf.data.Dataset, training data to be fed in tf.keras.Model.fit().
    validation_ds: tf.data.Dataset, validation data to be fed in
      tf.keras.Model.fit().
    steps_per_epoch: Integer or None. Total number of steps (batches of samples)
      before declaring one epoch finished and starting the next epoch. If
      `steps_per_epoch` is None, the epoch will run until the input dataset is
      exhausted.

  Returns:
    The tf.keras.callbacks.History object returned by tf.keras.Model.fit().
  """
  loss = tf.keras.losses.CategoricalCrossentropy(
      label_smoothing=hparams.label_smoothing)
  model.compile(
      optimizer=tf.keras.optimizers.SGD(
          lr=hparams.learning_rate, momentum=hparams.momentum),
      loss=loss,
      metrics=["accuracy"])

  return model.fit(
      train_ds,
      epochs=hparams.train_epochs,
      steps_per_epoch=steps_per_epoch,
      validation_data=validation_ds)


def train_model(model, hparams, train_ds, validation_ds, steps_per_epoch):
  """Trains model with the given data and hyperparameters.

  Args:
    model: The tf.keras.Model from _build_model().
    hparams: A namedtuple of hyperparameters. This function expects
      .train_epochs: a Python integer with the number of passes over the
        training dataset;
      .learning_rate: a Python float forwarded to the optimizer; Base learning
        rate when train batch size is 256. Linear to the batch size;
      .batch_size: a Python integer, the number of examples returned by each
        call to the generators;
      .warmup_steps: a Python integer, the number of warmup steps for warmup
        schedule on learning rate. If None, default warmup_steps is used;
      .model_dir: a Python string, the location of the model checkpoint files.
    train_ds: tf.data.Dataset, training data to be fed in tf.keras.Model.fit().
    validation_ds: tf.data.Dataset, validation data to be fed in
      tf.keras.Model.fit().
    steps_per_epoch: Integer or None. Total number of steps (batches of samples)
      before declaring one epoch finished and starting the next epoch. If
      `steps_per_epoch` is None, the epoch will run until the input dataset is
      exhausted.

  Returns:
    The tf.keras.callbacks.History object returned by tf.keras.Model.fit().
  """
  if steps_per_epoch is None:
    logging.info(
        "steps_per_epoch is None, use %d as the estimated steps_per_epoch",
        model_util.ESTIMITED_STEPS_PER_EPOCH)
    steps_per_epoch = model_util.ESTIMITED_STEPS_PER_EPOCH

  # Learning rate is linear to batch size.
  learning_rate = hparams.learning_rate * hparams.batch_size / 256

  # Gets decay steps.
  total_training_steps = steps_per_epoch * hparams.train_epochs
  default_decay_steps = DEFAULT_DECAY_SAMPLES // hparams.batch_size
  decay_steps = max(total_training_steps, default_decay_steps)

  warmup_steps = hparams.warmup_steps
  if warmup_steps is None:
    warmup_steps = DEFAULT_WARMUP_EPOCHS * steps_per_epoch
  optimizer = create_optimizer(learning_rate, decay_steps, warmup_steps)

  loss = tf.keras.losses.CategoricalCrossentropy(label_smoothing=0.1)
  model.compile(optimizer=optimizer, loss=loss, metrics=["accuracy"])

  summary_dir = os.path.join(hparams.model_dir, "summaries")
  summary_callback = tf.keras.callbacks.TensorBoard(summary_dir)
  # Save checkpoint every 20 epochs.
  checkpoint_path = os.path.join(hparams.model_dir, "checkpoint")
  checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
      checkpoint_path, save_weights_only=True, period=20)

  # Trains the models.
  return model.fit(
      train_ds,
      epochs=hparams.train_epochs,
      validation_data=validation_ds,
      callbacks=[summary_callback, checkpoint_callback])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import re

from tensorflow.core.framework import graph_pb2
from tensorflow.core.framework import node_def_pb2
from tensorflow.python.framework import tensor_util

# Custom op name for fused depthwise conv2d
FUSED_DEPTHWISE_CONV2D = 'FusedDepthwiseConv2dNative'
# The grappler op name for fused MatMul which starts with '_'
FUSED_MATMUL = '_FusedMatMul'

def node_from_map(node_map, name):
  """Pulls a node def from a dictionary for a given name.

  Args:
    node_map: Dictionary containing an entry indexed by name for every node.
    name: Identifies the node we want to find.

  Returns:
    NodeDef of the node with the given name.

  Raises:
    ValueError: If the node isn't present in the dictionary.
  """
  stripped_name = node_name_from_input(name)
  if stripped_name not in node_map:
    raise ValueError("No node named '%s' found in map." % name)
  return node_map[stripped_name]


def values_from_const(node_def):
  """Extracts the values from a const NodeDef as a numpy ndarray.

  Args:
    node_def: Const NodeDef that has the values we want to access.

  Returns:
    Numpy ndarray containing the values.

  Raises:
    ValueError: If the node isn't a Const.
  """
  if node_def.op != "Const":
    raise ValueError(
        "Node named '%s' should be a Const op for values_from_const." %
        node_def.name)
  input_tensor = node_def.attr["value"].tensor
  tensor_value = tensor_util.MakeNdarray(input_tensor)
  return tensor_value

# Whether to scale by gamma after normalization.
def scale_after_normalization(node):
  if node.op == "BatchNormWithGlobalNormalization":
    return node.attr["scale_after_normalization"].b
  return True

def node_name_from_input(node_name):
  """Strips off ports and other decorations to get the underlying node name."""
  if node_name.startswith("^"):
    node_name = node_name[1:]
  m = re.search(r"(.*):\d+$", node_name)
  if m:
    node_name = m.group(1)
  return node_name

def cleanup_graph_def(input_graph_def, nodes_to_skip, inputs_to_remove):
  """Clean up the graph def by removing the skipped nodes and clean up the nodes
    with inputs that have been removed.

  Args:
    input_graph_def: GraphDef object to be cleaned.
    node_to_skip: Dict with node names to be skipped.
    inputs_to_remove: List of nodes to be removed from inputs of all nodes.
  Returns:
    GraphDef that has been cleaned.

  """
  result_graph_def = graph_pb2.GraphDef()
  for node in input_graph_def.node:
    if node.name in nodes_to_skip:
      continue
    new_node = node_def_pb2.NodeDef()
    new_node.CopyFrom(node)
    for value in inputs_to_remove:
      for i, input_node in enumerate(new_node.input):
        if input_node == value.name:
          new_node.input[i] = value.input[0]
    result_graph_def.node.extend([new_node])
  result_graph_def.library.CopyFrom(input_graph_def.library)
  result_graph_def.versions.CopyFrom(input_graph_def.versions)
  return result_graph_def

def rename_constants(node_list, prefix):
  """Update all constants name by adding a prefix.

  Args:
    node_list: NodeDef list to update.
    prefix: string to add to the constant name.
  Returns:
    NodeDef list that has been updated.

  """
  nodes = []
  constant_names = [node.name for node in node_list if node.op == 'Const']
  for node in node_list:
    new_node = node_def_pb2.NodeDef()
    new_node.CopyFrom(node)
    nodes.append(new_node)
    if node.op == 'Const':
      new_node.name = prefix + '/' + node.name
    else:
      for i, input_node in enumerate(new_node.input):
        for name in constant_names:
          if input_node.startswith(name):
            new_node.input[i] = prefix + '/' + input_node
  return nodes



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Inspect tfrecord dataset."""
import os
from absl import app
from absl import flags
from absl import logging
import numpy as np
from PIL import Image
import tensorflow as tf

from tensorflow_examples.lite.model_maker.third_party.efficientdet import dataloader
from tensorflow_examples.lite.model_maker.third_party.efficientdet import hparams_config
from tensorflow_examples.lite.model_maker.third_party.efficientdet import utils
from tensorflow_examples.lite.model_maker.third_party.efficientdet.visualize import vis_utils

flags.DEFINE_string('save_samples_dir', 'tfrecord_samples',
                    'Location of samples to save')
flags.DEFINE_string('model_name', 'efficientdet-d0',
                    'model name for config and image_size')
flags.DEFINE_string(
    'hparams', '', 'Comma separated k=v pairs of hyperparameters or a module'
    ' containing attributes to use as hyperparameters.')
flags.DEFINE_integer('samples', 10,
                     'Number of random samples for visualization.')
flags.DEFINE_string('file_pattern', None,
                    'Glob for data files (e.g., COCO train - minival set)')
flags.DEFINE_bool('eval', False, 'flag for file pattern mode i.e eval')
FLAGS = flags.FLAGS


class RecordInspect:
  """Inspection Class."""

  def __init__(self, config):
    """Initializes RecordInspect with passed config.

    Args:
        config: config file to initialize input_fn.
    """
    self.input_fn = dataloader.InputReader(
        FLAGS.file_pattern,
        is_training=not FLAGS.eval,
        use_fake_data=False,
        max_instances_per_image=config.max_instances_per_image)

    self.params = dict(
        config.as_dict(), batch_size=FLAGS.samples, model_name=FLAGS.model_name)
    logging.info(self.params)
    self.cls_to_label = config.label_map
    os.makedirs(FLAGS.save_samples_dir, exist_ok=True)

  def visualize(self):
    """save tfrecords images with bounding boxes."""
    vis_ds = self.input_fn(params=self.params)
    data = next(iter(vis_ds))  # iterable.
    images = data[0]
    gt_data = data[1]['groundtruth_data']

    # scales
    scale_to_org = data[1]['image_scales']
    scales = 1.0 / scale_to_org
    offset = tf.constant([0.485, 0.456, 0.406])
    offset = tf.reshape(offset, (1, 1, -1))
    scale_image = tf.constant([0.229, 0.224, 0.225])
    scale_image = tf.reshape(scale_image, (1, 1, -1))

    logging.info('Visualizing TfRecords %s', FLAGS.file_pattern)
    for i, zip_data in enumerate(zip(gt_data, images, scales)):
      gt, image, scale = zip_data
      boxes = gt[:, :4]
      boxes = boxes[np.any(boxes > 0, axis=1)].numpy()
      if boxes.shape[0] > 0:
        classes = gt[:boxes.shape[0], -1].numpy()
        try:
          category_index = {idx: {'id': idx, 'name': self.cls_to_label[idx]}
                            for idx in np.asarray(classes, dtype=np.int)}
        except Exception:  # pylint: disable=broad-except
          category_index = {}

        # unnormalize image.
        image *= scale_image
        image += offset

        # 0-255. range
        image = np.asarray(image.numpy() * 255., dtype=np.uint8)

        # scale to image_size
        boxes *= scale.numpy()

        image = vis_utils.visualize_boxes_and_labels_on_image_array(
            image,
            boxes=boxes,
            classes=classes.astype(int),
            scores=np.ones(boxes.shape[0]),
            category_index=category_index,
            line_thickness=2,
            skip_scores=True)
        image = Image.fromarray(image)
        image.save(os.path.join(FLAGS.save_samples_dir, f'sample{i}.jpg'))


def main(_):
  # Parse and override hparams
  config = hparams_config.get_detection_config(FLAGS.model_name)
  config.override(FLAGS.hparams)

  # Parse image size in case it is in string format.
  config.image_size = utils.parse_image_size(config.image_size)
  try:
    recordinspect = RecordInspect(config)
    recordinspect.visualize()
  except Exception as e:  # pylint: disable=broad-except
    logging.error(e)
  else:
    logging.info('Done, please find samples at %s', FLAGS.save_samples_dir)


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""CLI utility to generate APIs.

Usage:
python api_gen.py --output_dir=<path to output> --version=0.x.x \
  --base_package=tflite_model_maker
"""

import argparse
import json
import pathlib
from typing import Dict, Sequence

from tensorflow_examples.lite.model_maker.core.api import api_util
from tensorflow_examples.lite.model_maker.core.api import deprecated_api
from tensorflow_examples.lite.model_maker.core.api import golden_api_doc


def parse_arguments():
  """Parse arguments for API gen."""
  parser = argparse.ArgumentParser()
  parser.add_argument(
      '-o', '--output_dir', type=str, help='Base dir to output generated APIs.')
  parser.add_argument(
      '-v',
      '--version',
      type=str,
      default='0.0.0dev',
      help='Version of the package.')
  parser.add_argument(
      '-b',
      '--base_package',
      type=str,
      default='tflite_model_maker',
      help='Version of the package.')
  return parser.parse_args()


def _read_golden_text(json_file: str) -> str:
  """Reads Golden file as text."""
  path = pathlib.Path(__file__).with_name(json_file)
  with open(path) as f:
    return f.read()


def load_golden(json_file: str) -> Dict[str, Sequence[str]]:
  """Loads Golden APIs."""
  content = _read_golden_text(json_file)
  return json.loads(content)


def run(output_dir: str, base_package: str, version: str) -> None:
  """Runs main."""
  imports = load_golden('golden_api.json')
  imports_doc = golden_api_doc.DOCS
  deprecated_imports = deprecated_api.IMPORTS
  api_util.write_packages(
      output_dir,
      imports,
      imports_doc,
      base_package,
      version,
      deprecated_imports=deprecated_imports)


def main() -> None:
  args = parse_arguments()
  run(args.output_dir, args.base_package, args.version)


if __name__ == '__main__':
  main()



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A few predefined label id mapping."""
import tensorflow as tf
import yaml
from tensorflow_examples.lite.model_maker.third_party.efficientdet import hparams_config

coco = {
    # 0: 'background',
    1: 'person',
    2: 'bicycle',
    3: 'car',
    4: 'motorcycle',
    5: 'airplane',
    6: 'bus',
    7: 'train',
    8: 'truck',
    9: 'boat',
    10: 'traffic light',
    11: 'fire hydrant',
    13: 'stop sign',
    14: 'parking meter',
    15: 'bench',
    16: 'bird',
    17: 'cat',
    18: 'dog',
    19: 'horse',
    20: 'sheep',
    21: 'cow',
    22: 'elephant',
    23: 'bear',
    24: 'zebra',
    25: 'giraffe',
    27: 'backpack',
    28: 'umbrella',
    31: 'handbag',
    32: 'tie',
    33: 'suitcase',
    34: 'frisbee',
    35: 'skis',
    36: 'snowboard',
    37: 'sports ball',
    38: 'kite',
    39: 'baseball bat',
    40: 'baseball glove',
    41: 'skateboard',
    42: 'surfboard',
    43: 'tennis racket',
    44: 'bottle',
    46: 'wine glass',
    47: 'cup',
    48: 'fork',
    49: 'knife',
    50: 'spoon',
    51: 'bowl',
    52: 'banana',
    53: 'apple',
    54: 'sandwich',
    55: 'orange',
    56: 'broccoli',
    57: 'carrot',
    58: 'hot dog',
    59: 'pizza',
    60: 'donut',
    61: 'cake',
    62: 'chair',
    63: 'couch',
    64: 'potted plant',
    65: 'bed',
    67: 'dining table',
    70: 'toilet',
    72: 'tv',
    73: 'laptop',
    74: 'mouse',
    75: 'remote',
    76: 'keyboard',
    77: 'cell phone',
    78: 'microwave',
    79: 'oven',
    80: 'toaster',
    81: 'sink',
    82: 'refrigerator',
    84: 'book',
    85: 'clock',
    86: 'vase',
    87: 'scissors',
    88: 'teddy bear',
    89: 'hair drier',
    90: 'toothbrush',
}

voc = {
    # 0: 'background',
    1: 'aeroplane',
    2: 'bicycle',
    3: 'bird',
    4: 'boat',
    5: 'bottle',
    6: 'bus',
    7: 'car',
    8: 'cat',
    9: 'chair',
    10: 'cow',
    11: 'diningtable',
    12: 'dog',
    13: 'horse',
    14: 'motorbike',
    15: 'person',
    16: 'pottedplant',
    17: 'sheep',
    18: 'sofa',
    19: 'train',
    20: 'tvmonitor',
}

waymo = {
    # 0: 'background',
    1: 'vehicle',
    2: 'pedestrian',
    3: 'cyclist',
}


def get_label_map(mapping):
  """Get label id map based on the name, filename, or dict."""
  # case 1: if it is None or dict, just return it.
  if not mapping or isinstance(mapping, dict):
    return mapping

  if isinstance(mapping, hparams_config.Config):
    return mapping.as_dict()

  # case 2: if it is a yaml file, load it to a dict and return the dict.
  assert isinstance(mapping, str), 'mapping must be dict or str.'
  if mapping.endswith('.yaml'):
    with tf.io.gfile.GFile(mapping) as f:
      return yaml.load(f, Loader=yaml.FullLoader)

  # case 3: it is a name of a predefined dataset.
  return {'coco': coco, 'voc': voc, 'waymo': waymo}[mapping]



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
r"""Convert PASCAL dataset to TFRecord.

Example usage:
    python create_pascal_tfrecord.py  --data_dir=/tmp/VOCdevkit  \
        --year=VOC2012  --output_path=/tmp/pascal
"""
import hashlib
import io
import json
import os

from absl import app
from absl import flags
from absl import logging

from lxml import etree
import PIL.Image
import tensorflow as tf

from tensorflow_examples.lite.model_maker.third_party.efficientdet.dataset import tfrecord_util

FLAGS = flags.FLAGS

SETS = ['train', 'val', 'trainval', 'test']
YEARS = ['VOC2007', 'VOC2012', 'merged']

pascal_label_map_dict = {
    'background': 0,
    'aeroplane': 1,
    'bicycle': 2,
    'bird': 3,
    'boat': 4,
    'bottle': 5,
    'bus': 6,
    'car': 7,
    'cat': 8,
    'chair': 9,
    'cow': 10,
    'diningtable': 11,
    'dog': 12,
    'horse': 13,
    'motorbike': 14,
    'person': 15,
    'pottedplant': 16,
    'sheep': 17,
    'sofa': 18,
    'train': 19,
    'tvmonitor': 20,
}


def define_flags():
  """Define the flags."""
  flags.DEFINE_string('data_dir', '',
                      'Root directory to raw PASCAL VOC dataset.')
  flags.DEFINE_string('set', 'train', 'Convert training set, validation set or '
                      'merged set.')
  flags.DEFINE_string('annotations_dir', 'Annotations',
                      '(Relative) path to annotations directory.')
  flags.DEFINE_string('year', 'VOC2007', 'Desired challenge year.')
  flags.DEFINE_string('output_path', '', 'Path to output TFRecord and json.')
  flags.DEFINE_string('label_map_json_path', None,
                      'Path to label map json file with a dictionary.')
  flags.DEFINE_boolean('ignore_difficult_instances', False, 'Whether to ignore '
                       'difficult instances')
  flags.DEFINE_integer('num_shards', 100, 'Number of shards for output file.')
  flags.DEFINE_integer('num_images', None, 'Max number of imags to process.')


class UniqueId(object):
  """Class to get the unique {image/ann}_id each time calling the functions."""

  def __init__(self):
    self.image_id = 0
    self.ann_id = 0

  def get_image_id(self):
    self.image_id += 1
    return self.image_id

  def get_ann_id(self):
    self.ann_id += 1
    return self.ann_id


def dict_to_tf_example(data,
                       images_dir,
                       label_map_dict,
                       unique_id,
                       ignore_difficult_instances=False,
                       ann_json_dict=None):
  """Convert XML derived dict to tf.Example proto.

  Notice that this function normalizes the bounding box coordinates provided
  by the raw data.

  Args:
    data: dict holding PASCAL XML fields for a single image (obtained by running
      tfrecord_util.recursive_parse_xml_to_dict)
    images_dir: Path to the directory holding raw images.
    label_map_dict: A map from string label names to integers ids.
    unique_id: UniqueId object to get the unique {image/ann}_id for the image
      and the annotations.
    ignore_difficult_instances: Whether to skip difficult instances in the
      dataset  (default: False).
    ann_json_dict: annotation json dictionary.

  Returns:
    example: The converted tf.Example.

  Raises:
    ValueError: if the image pointed to by data['filename'] is not a valid JPEG
  """
  full_path = os.path.join(images_dir, data['filename'])
  with tf.io.gfile.GFile(full_path, 'rb') as fid:
    encoded_jpg = fid.read()
  encoded_jpg_io = io.BytesIO(encoded_jpg)
  image = PIL.Image.open(encoded_jpg_io)
  if image.format != 'JPEG':
    raise ValueError('Image format not JPEG')
  key = hashlib.sha256(encoded_jpg).hexdigest()

  image_id = unique_id.get_image_id()

  width = int(data['size']['width'])
  height = int(data['size']['height'])
  if ann_json_dict:
    image = {
        'file_name': data['filename'],
        'height': height,
        'width': width,
        'id': image_id,
    }
    ann_json_dict['images'].append(image)

  xmin = []
  ymin = []
  xmax = []
  ymax = []
  area = []
  classes = []
  classes_text = []
  truncated = []
  poses = []
  difficult_obj = []
  if 'object' in data:
    for obj in data['object']:
      difficult = bool(int(obj['difficult']))
      if ignore_difficult_instances and difficult:
        continue

      difficult_obj.append(int(difficult))

      xmin.append(float(obj['bndbox']['xmin']) / width)
      ymin.append(float(obj['bndbox']['ymin']) / height)
      xmax.append(float(obj['bndbox']['xmax']) / width)
      ymax.append(float(obj['bndbox']['ymax']) / height)
      area.append((xmax[-1] - xmin[-1]) * (ymax[-1] - ymin[-1]))
      classes_text.append(obj['name'].encode('utf8'))
      classes.append(label_map_dict[obj['name']])
      truncated.append(int(obj['truncated']))
      poses.append(obj['pose'].encode('utf8'))

      if ann_json_dict:
        abs_xmin = int(obj['bndbox']['xmin'])
        abs_ymin = int(obj['bndbox']['ymin'])
        abs_xmax = int(obj['bndbox']['xmax'])
        abs_ymax = int(obj['bndbox']['ymax'])
        abs_width = abs_xmax - abs_xmin
        abs_height = abs_ymax - abs_ymin
        ann = {
            'area': abs_width * abs_height,
            'iscrowd': 0,
            'image_id': image_id,
            'bbox': [abs_xmin, abs_ymin, abs_width, abs_height],
            'category_id': label_map_dict[obj['name']],
            'id': unique_id.get_ann_id(),
            'ignore': 0,
            'segmentation': [],
        }
        ann_json_dict['annotations'].append(ann)

  example = tf.train.Example(
      features=tf.train.Features(
          feature={
              'image/height':
                  tfrecord_util.int64_feature(height),
              'image/width':
                  tfrecord_util.int64_feature(width),
              'image/filename':
                  tfrecord_util.bytes_feature(data['filename'].encode('utf8')),
              'image/source_id':
                  tfrecord_util.bytes_feature(str(image_id).encode('utf8')),
              'image/key/sha256':
                  tfrecord_util.bytes_feature(key.encode('utf8')),
              'image/encoded':
                  tfrecord_util.bytes_feature(encoded_jpg),
              'image/format':
                  tfrecord_util.bytes_feature('jpeg'.encode('utf8')),
              'image/object/bbox/xmin':
                  tfrecord_util.float_list_feature(xmin),
              'image/object/bbox/xmax':
                  tfrecord_util.float_list_feature(xmax),
              'image/object/bbox/ymin':
                  tfrecord_util.float_list_feature(ymin),
              'image/object/bbox/ymax':
                  tfrecord_util.float_list_feature(ymax),
              'image/object/area':
                  tfrecord_util.float_list_feature(area),
              'image/object/class/text':
                  tfrecord_util.bytes_list_feature(classes_text),
              'image/object/class/label':
                  tfrecord_util.int64_list_feature(classes),
              'image/object/difficult':
                  tfrecord_util.int64_list_feature(difficult_obj),
              'image/object/truncated':
                  tfrecord_util.int64_list_feature(truncated),
              'image/object/view':
                  tfrecord_util.bytes_list_feature(poses),
          }))
  return example


def main(_):
  if FLAGS.set not in SETS:
    raise ValueError('set must be in : {}'.format(SETS))
  if FLAGS.year not in YEARS:
    raise ValueError('year must be in : {}'.format(YEARS))
  if not FLAGS.output_path:
    raise ValueError('output_path cannot be empty.')

  data_dir = FLAGS.data_dir
  years = ['VOC2007', 'VOC2012']
  if FLAGS.year != 'merged':
    years = [FLAGS.year]

  output_dir = os.path.dirname(FLAGS.output_path)
  if not tf.io.gfile.exists(output_dir):
    tf.io.gfile.makedirs(output_dir)
  logging.info('Writing to output directory: %s', output_dir)

  writers = [
      tf.io.TFRecordWriter(FLAGS.output_path + '-%05d-of-%05d.tfrecord' %
                           (i, FLAGS.num_shards))
      for i in range(FLAGS.num_shards)
  ]

  if FLAGS.label_map_json_path:
    with tf.io.gfile.GFile(FLAGS.label_map_json_path, 'rb') as f:
      label_map_dict = json.load(f)
  else:
    label_map_dict = pascal_label_map_dict

  ann_json_dict = {
      'images': [],
      'type': 'instances',
      'annotations': [],
      'categories': []
  }

  unique_id = UniqueId()
  for year in years:
    example_class = list(label_map_dict.keys())[1]
    examples_path = os.path.join(data_dir, year, 'ImageSets', 'Main',
                                 example_class + '_' + FLAGS.set + '.txt')
    examples_list = tfrecord_util.read_examples_list(examples_path)
    annotations_dir = os.path.join(data_dir, year, FLAGS.annotations_dir)

    for class_name, class_id in label_map_dict.items():
      cls = {'supercategory': 'none', 'id': class_id, 'name': class_name}
      ann_json_dict['categories'].append(cls)

    logging.info('Reading from PASCAL %s dataset.', year)
    for idx, example in enumerate(examples_list):
      if FLAGS.num_images and idx >= FLAGS.num_images:
        break
      if idx % 100 == 0:
        logging.info('On image %d of %d', idx, len(examples_list))
      path = os.path.join(annotations_dir, example + '.xml')
      with tf.io.gfile.GFile(path, 'r') as fid:
        xml_str = fid.read()
      xml = etree.fromstring(xml_str)
      data = tfrecord_util.recursive_parse_xml_to_dict(xml)['annotation']

      img_dir = os.path.join(FLAGS.data_dir, data['folder'], 'JPEGImages')

      tf_example = dict_to_tf_example(
          data,
          img_dir,
          label_map_dict,
          unique_id,
          FLAGS.ignore_difficult_instances,
          ann_json_dict=ann_json_dict)
      writers[idx % FLAGS.num_shards].write(tf_example.SerializeToString())

  for writer in writers:
    writer.close()

  json_file_path = os.path.join(
      os.path.dirname(FLAGS.output_path),
      'json_' + os.path.basename(FLAGS.output_path) + '.json')
  with tf.io.gfile.GFile(json_file_path, 'w') as f:
    json.dump(ann_json_dict, f)


if __name__ == '__main__':
  define_flags()
  app.run(main)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
r"""Simple tools for TensorRT.

Example usage:

$ export ROOT=/tmp/d4
$ python model_inspect.py --runmode=freeze --model_name=efficientdet-d4 \
    --logdir=$ROOT  # --hparams=xyz.yaml
$ python tensorrt.py --tf_savedmodel_dir=$ROOT/savedmodel \
    --trt_savedmodel_dir=$ROOT/trtmodel
"""
import time
from absl import app
from absl import flags
import numpy as np
import tensorflow.compat.v1 as tf
# pylint: disable=g-direct-tensorflow-import
from tensorflow.python.compiler.tensorrt import trt_convert as trt

flags.DEFINE_string('tf_savedmodel_dir', None, 'TensorFlow saved model dir.')
flags.DEFINE_string('trt_savedmodel_dir', None, 'TensorRT saved model dir.')
FLAGS = flags.FLAGS


def convert2trt(tf_savedmodel_dir: str, trt_savedmodel_dir: str):
  converter = trt.TrtGraphConverter(
      input_saved_model_dir=tf_savedmodel_dir,
      max_workspace_size_bytes=(2 << 20),
      precision_mode='FP16',
      maximum_cached_engines=1)
  converter.convert()
  converter.save(trt_savedmodel_dir)


def benchmark(trt_savedmodel_dir: str, warmup_runs: int = 5, bm_runs: int = 20):
  """Benchmark TRT latency for a given TRT saved model."""
  with tf.Session() as sess:
    # First load the Saved Model into the session
    tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING],
                               trt_savedmodel_dir)
    graph = tf.get_default_graph()
    input_shape = graph.get_tensor_by_name('input:0').shape
    x = np.ones(input_shape).astype(np.float32)
    ss = lambda i: '' if i == 0 else '_%d' % i
    outputs = ['box_net/box-predict%s/BiasAdd:0' % ss(i) for i in range(1)]
    outputs += ['class_net/class-predict%s/BiasAdd:0' % ss(i) for i in range(5)]
    # Apply reduce_sum to avoid massive data move between GPU and CPU.
    outputs = [tf.reduce_sum(graph.get_tensor_by_name(i)) for i in outputs]

    # warmup
    for _ in range(warmup_runs):
      sess.run(outputs, feed_dict={'input:0': x})
    # benchmark
    s = time.perf_counter()
    for _ in range(bm_runs):
      sess.run(outputs, feed_dict={'input:0': x})
    e = time.perf_counter()
    print('Benchmark latency=%.4f  FPS=%.2f', (e - s) / bm_runs,
          bm_runs / (e - s))


def main(_):
  if FLAGS.tf_savedmodel_dir:
    convert2trt(FLAGS.tf_savedmodel_dir, FLAGS.trt_savedmodel_dir)
  benchmark(FLAGS.trt_savedmodel_dir, FLAGS.warmup_runs, FLAGS.bm_runs)


if __name__ == '__main__':
  flags.mark_flag_as_required('trt_savedmodel_dir')
  tf.disable_v2_behavior()
  app.run(main)
